module Data.Enumerable
	( Enumerable
	, count
	, select
	, withdraw
	, enumerate
	, isNil

	, rangeEnum
	, rangeNum

	, roll
	, draw
	, drawRandom

	, from
	, listOf
	, listRangeOf

	, addOrNot
	, permutations
	, addOrNotA
	, permutationsA

	, drawsOfN
	, drawsOfNRange
	) where

import Control.Applicative

import Utility.Rand

data Enumerable a
	= Enumerable
		{ _count :: Integer
		, selector :: Integer -> a
		}

count :: Enumerable a -> Integer
count e = _count e

instance Functor Enumerable where
	fmap f (Enumerable c g) = Enumerable c $ f . g

instance Semigroup (Enumerable a) where
	(<>) = (<|>)

instance Monoid (Enumerable a) where
	mempty = Enumerable 0 $ \_ -> error "empty enumerable value"

-- this applictive indexing is maximally-distinct: if there are c values in a deck, then in the combined deck you won't see a repeat of the values until viewing c+1 items, regardless of the other values. this trait sometimes persists across multiple applications of <*>
-- one after the other
instance Applicative Enumerable where
	pure v = Enumerable 1 $ const v
	(Enumerable c f) <*> (Enumerable c' g) =
		Enumerable (c * c') $ \i -> let
				(%) = mod
				fact = gcd c c'
				loop = c * c' `div` fact
				o = i `div` loop
-- original enumeration: (1,1), (1,2), (1,3), (2,1), ...
--				j = i % c
--				k = i `div` c
-- this enumeration gets more-distinct values: (1,1), (2,2), (3,3), (1,2), (2,3), ...
				j = i % c
				k = ((i % c) + (i `div` c) * (c' + fact) + o) % c'
			in f j $ g k

-- one or the other
instance Alternative Enumerable where
	empty = Enumerable 0 $ \_ -> error "empty enumerable value"
	a <|> b = interleave a b
	{- Enumerable (c + c') $ \i -> case i - c of
		x | x >= 0 -> v' x
			| otherwise -> v i
	EnumerableNil <|> f = f
	f <|> EnumerableNil = f
	-}

-- interleave at a 1:x ratio until one deck runs out of values, and then list the rest. could be improved; see below
interleave :: Enumerable a -> Enumerable a -> Enumerable a
interleave (Enumerable 0 v) b = b
interleave a (Enumerable 0 v) = a
interleave (Enumerable c v) (Enumerable c' v') =
	let
		largerV = if c > c' then v else v'
		smallerV = if c > c' then v' else v
		largerC = max c c'
		smallerC = min c c'

		sharedTil = (smallerC * dealsPerIteration)
		lis = largerC `div` smallerC
		sis = 1
		rems = largerC `mod` smallerC
		dealsPerIteration = lis + sis
		lOffset = (-sis)
	in Enumerable (c + c') $ \i -> if i >= sharedTil
		then largerV $ (i - smallerC)
		else case i `mod` dealsPerIteration of -- so 0 1 2 3 4
			0 -> smallerV $ i `div` dealsPerIteration
			i' -> let onIteration = i `div` dealsPerIteration
				in largerV $ (onIteration * lis) + (i' + lOffset)
{-
		\i ->
			if i < sharedTil
				then if (i `mod` 2) == 0
					then v $ i `div` 2
					else v' $ i `div` 2
				else largerV (i - sharedIndices)
-}
{-

the current version will always generate a listing like `x (1, y) + 1 (0, z)`. this is fine in most cases, but when the deck sizes share a factor (or something?) it can lead to some suboptimal splits:
	-- current, dumb version:
	(20, 30) -> 20 ( 1,  1) + 1 (0,10) -- (30 `div` 20 = 1; 30 `mod` 20 = 10)
	-- ENLIGHTENED version:
	(20, 30) -> 10 ( 2,  3)

... to enumerate the values we basically want to find a shared factor, or smth close to it
-}

{-
instance Monad Enumerable where
	return = pure
	(Enumerable c v) >>= f = ...? the thing is we don't really know how the output varies w/ the input. so we MIGHT be like, oh, just evaluate every `i` and then wrap the resulting value up so we have a bunch of Enumerables that we can sequence. but then we have to evaluate every `i`, which is not practical. additionally, we don't know if those variations actually matter: f >>= return == f, but here we'd end up with a very different value. consequently, i don't think it's possible for this to be an Monad instance, even if it is maybe a monad conceptually. or at least, it's certainly not going to be implemented in this fashion.
	_ >>= _ = EnumerableNil

monad laws:
	return a >>= f    ≡ f a
	m >>= return      ≡ m
	(f >=> g) >=> g   ≡ f >=> (g >=> h)
		(m >>= f) >>= g ≡ m >>= (\x -> f x >>= g)

instance MonadPlus Enumerable where
	mzero = EnumerableNil
	mplus = (<|>)
-}

isNil :: Enumerable a -> Bool
isNil (Enumerable 0 _) = True
isNil _ = False

roll :: RandomGen g => Enumerable a -> Rand g a
roll enum = do
	i <- liftRand $ randomR (0, _count enum-1)
	case select enum i of
		Nothing -> error $ "incoherent enumerable value"
		Just x -> return x

drawRandom :: RandomGen g => Enumerable a -> Rand g (a, Enumerable a)
drawRandom enum = do
	i <- liftRand $ randomR (0, count enum-1)
	return $ draw enum i

draw :: Enumerable a -> Integer -> (a, Enumerable a)
draw enum i =
	case select enum i of
		Nothing -> error $ "incoherent enumerable value"
		Just x -> (x, withdraw enum i)

withdraw :: Enumerable a -> Integer -> Enumerable a
withdraw (Enumerable c s) i
	| c <= 0 = error "empty enumerable value"
	| otherwise = Enumerable (c-1) $ \i' -> if i' < i then s i' else s (i' + 1)

select :: Enumerable a -> Integer -> Maybe a
select enum i
	| i < 0 || i >= _count enum = Nothing
	| otherwise = Just $ selector enum i

enumerate :: Enumerable a -> [a]
enumerate (Enumerable c s)
	| c <= 0 = []
	| otherwise = s <$> [0..c-1]

-- different ways of specifying a range of values. it's assumed that these atomic elements will have counts less than Int's maxBound and thus can be constructed like this safely.
rangeEnum :: Enum a => (a, a) -> Enumerable a
rangeEnum (lo, hi) = Enumerable (fromIntegral all) $ \i -> toEnum $ fromEnum lo + fromIntegral i
	where
		all = fromEnum hi - fromEnum lo + 1

rangeNum :: Integral a => (a, a) -> Enumerable a
rangeNum (lo, hi) = Enumerable (fromIntegral all) $ \i -> lo + fromIntegral i
	where
		all = hi - lo + 1

-- this does require evaluating the entire input list, so don't make it too huge
from :: [a] -> Enumerable a
from [] = Enumerable 0 $ \_ -> error "empty enumerable value"
from vs = Enumerable (fromIntegral $ length vs) $ \i -> vs !! fromIntegral i


listOf :: Int -> Enumerable a -> Enumerable [a]
listOf 0 _ = pure []
listOf 1 v = fmap pure v
listOf n v | n < 0 = error "listOf: cannot make zero-length list"
	| otherwise = (:) <$> v <*> listOf (n - 1) v

listRangeOf :: (Int, Int) -> Enumerable a -> Enumerable [a]
listRangeOf (lo, hi) v
	| lo == hi = listOf lo v
	| lo > hi = error "listOfRange: low end of range cannot be larger than high end"
	| otherwise = listOf lo v `mappend` listRangeOf (lo+1, hi) v


permutationsA :: [Enumerable (a -> a)] -> Enumerable a -> Enumerable a
permutationsA opts v = foldr addOrNotA v opts

addOrNotA :: Enumerable (a -> a) -> Enumerable a -> Enumerable a
addOrNotA f a = a <> (f <*> a)

permutations :: [a -> a] -> Enumerable a -> Enumerable a
permutations opts v = foldr addOrNot v opts

addOrNot :: (a -> a) -> Enumerable a -> Enumerable a
addOrNot opt v = v <> (opt <$> v)


mapWithIndex :: (Integer -> a -> b) -> Enumerable a -> Enumerable b
mapWithIndex f (Enumerable c g) = Enumerable c $ \i -> f i (g i)

oneCard :: Enumerable a -> Enumerable (a, Enumerable a)
oneCard vs = mapWithIndex (\i v -> (v, withdraw vs i)) vs

-- TODO: should maybe return a nested Enumerable? that would undo the work of making this bind fast, though
drawsOfN :: Int -> Enumerable a -> Enumerable [a]
drawsOfN 0 _ = pure []
drawsOfN 1 v = fmap pure v
drawsOfN n v | n < 0 = error "drawsOfN: cannot make < 0-length list"
  | otherwise = (\(h, rs) -> (h :) <$> drawsOfN (n-1) rs)
      `fixedBind` oneCard v

drawsOfNRange :: (Int, Int) -> Enumerable a -> Enumerable [a]
drawsOfNRange (lo, hi) v
  | lo == hi = drawsOfN lo v
  | lo > hi = error "distinctRangeListOf: low end of range can't be higher than high end"
  | otherwise = drawsOfN lo v <|> drawsOfNRange (lo+1, hi) v





-- because this is only used with oneCard, we know that every nested possibility space will have the same size, so instead of needing to evaluate all of them to check we can just check the first one and do an applicative-style deck multiplication
fixedJoin :: Enumerable (Enumerable a) -> Enumerable a
fixedJoin (Enumerable c f) = let
    c' = count $ f 0
  in Enumerable (c * c') $ \i -> let
      j = i `div` c'
      k = i `mod` c'
      (Enumerable _ g) = f j
    in g k
-- this is a (hugely) faster bind for a specific kind of nesting. specifically, this is only safe to use when the resultant probability space has the same number of output values for each subdeck. if that is not true, it will corrupt your space data.
fixedBind :: (a -> Enumerable b) -> Enumerable a -> Enumerable b
fixedBind f e = fixedJoin $ f <$> e
