module Description.Describe
	( describe
	, stub
	) where

import Control.Arrow (first, second)
import Data.Maybe (catMaybes)
import Data.List (sortBy, groupBy)
import Data.Ord (comparing)
import Data.Function (on)
import Data.Tree (Tree)

import Description.Pronoun (Gender)
import Description.TfNode (TfNode)
import Description.Build (build)
import Description.Construction (parse, Sentence(..))
import Description.Prose (prose, stubNoun, DescOrder(..))

-- returns: (errors, description lines)
describe :: Gender -> Gender -> Tree TfNode -> ([String], [String])
describe  n p tree =
	let
		(ws, ss) = prose $ build n p tree
		gss = if fewerThanXCockAccentLines 2 ss
			then groupBy (sameParagraph `on` label) $ sortBy (comparing label) $ ss
			else groupBy (separateCockParagraphs `on` label) $ sortBy (cockCompare `on` label) $ ss
	in
		  second (fmap unwords)
		. extract
		. fmap catMaybes
			$ parse ws <$$> gss

stub :: Gender -> Tree TfNode -> String
stub n tree = stubNoun n $ build n n tree

label :: Sentence b c a -> b
label (S b _ _) = b
label (PJ s _ _) = label s

(<$$>) :: Functor f => (a -> b) -> f (f a) -> f (f b)
(<$$>) = fmap . fmap

extract :: [[([a], b)]] -> ([a], [[b]])
extract = foldr f ([], [])
	where
		f :: [([a], b)] -> ([a], [[b]]) -> ([a], [[b]])
		f vs (as, bs) =
			let (as', bs') = first concat . unzip $ vs
			in (as' ++ as, bs' : bs)

sameParagraph :: DescOrder -> DescOrder -> Bool
sameParagraph a b = (elem a <$> ps) == (elem b <$> ps)
	where
		ps =
			[ [Summary, BodyBaseMaterial, BodyMaterial, BodyAccent, LimbDetail]
			, [HeadCoveringLine, HeadDetail, TongueLine, HornLine, HaloLine]
			, [CockSoft, CockSoftAccent
				, CockReveal, CockRevealAccent
				, CockHard, CockHardAccent, CockHardSize
				, CockCumAccent, BallCumAccent, BallAccent
				]
			]



cockCompare :: DescOrder -> DescOrder -> Ordering
cockCompare a b = compare (length $ dropWhile (/= b) vs) (length $ dropWhile (/= a) vs)
	where vs =
		[ Summary, BodyBaseMaterial, BodyMaterial, BodyAccent, LimbDetail
		, HeadCoveringLine, HeadDetail, TongueLine, HornLine, HaloLine
		, CockSoft, CockSoftAccent
		, CockReveal, CockRevealAccent
		, CockHardSize, CockHardAccent
		, CockCumAccent, BallCumAccent
		, BallAccent
		]

separateCockParagraphs :: DescOrder -> DescOrder -> Bool
separateCockParagraphs a b = (elem a <$> ps) == (elem b <$> ps)
	where
		ps =
			[ [Summary, BodyBaseMaterial, BodyMaterial, BodyAccent, LimbDetail]
			, [HeadCoveringLine, HeadDetail]
			, [TongueLine, HornLine, HaloLine]
			, [CockSoft, CockSoftAccent]
			, [CockReveal, CockRevealAccent]
			, [CockHardSize, CockHardAccent]
			, [CockCumAccent, BallCumAccent, BallAccent]
			]

fewerThanXCockAccentLines :: Int -> [Sentence DescOrder b c] -> Bool
fewerThanXCockAccentLines limit s = (limit >) $ length
	$ filter (\s -> label s `elem` [CockSoftAccent, CockHardAccent, CockRevealAccent])
		$ s
