module Description.Parse
	( parse

	, SymSem(..)
	) where

import Prelude hiding (Word, (/))

import Data.Char (toUpper, toLower)
import Data.Either
import Data.List
import Data.Maybe

import Control.Arrow (second)
import Control.Monad

import Description.Parse.Types
import Description.Dictionary
import Description.Pronoun

-- this is what attaches periods to the ends of sentences (which will also end up removing dangling punctuation if applicable)
parse :: [WordInfo] -> Sentence def Semantic String -> Maybe ([String], String)
parse ws sent = display . (++ [Punc Period]) <$> extract ws sent

extract :: [WordInfo] -> Sentence def Semantic String -> Maybe [Literal String]
extract ws (S _ (Diagram diagram) fs) = Just $ concat $ mapMaybe (render ws fs) $ diagram
-- extract ws (S _ (Diagram diagram) fs) = fmap concat $ render ws fs `mapM` diagram
extract ws (PJ sent1 prepFrag sent2) =
	extract ws sent1
	<> case extract ws sent2 of
		Just o -> Just $ Punc Comma : fromMaybe [] (render ws [] prepFrag) ++ o
		Nothing -> Just []

render :: [WordInfo] -> [Fragment Semantic String] -> Fragment Semantic String -> Maybe [Literal String]
render ws fs (Fragment sem word) =
	wordParse ws fs (if Verb `matches` sem
		then subjWord sem
		else getMatching sem ws) word
	where
		subjWord :: Semantic -> WordInfo
		subjWord sem = fromMaybe (WI 1 sem Nothing)
			$ (\(Fragment sem' word) -> getMatching sem' ws) <$> grab Subject fs

grab :: SymSem sem => sem -> [Fragment Semantic word] -> Maybe (Fragment Semantic word)
grab sem [] = Nothing
grab sem (f@(Fragment sem' w):rs) = if sem `matches` sem'
	then Just f
	else grab sem rs

-- we'd need to check a/an via a backwards traversal ('an' if the next word starts with a e i o u ho, 'a' otherwise)
display :: [Literal String] -> ([String], String)
display vs = second (concat . snd . mapAccumL punctuate (True, False)) $ partitionEithers $ foo <$> resolve Nothing vs
	where
		foo :: Literal String -> Either String (Literal String)
		foo AAn = Right (Stem "a") -- i think by this point all the AAns would've been parsed out, but, w/e
		foo (Stem a) = Right (Stem a)
		foo (Punc p) = Right (Punc p)
		foo (Error a) = Left a
		foo (Inflect _) = Left "inflect in wrong place :("
		foo Nil = Left "nil in wrong place :("

punctuate :: (Bool, Bool) -> Literal String -> ((Bool, Bool), String)
punctuate (toCap, leadingSpace) (Stem a) = ((False, True),
	(if leadingSpace then (' ' :) else id)
		$ if toCap then capitalize a else a)
-- don't know for sure if all of these punctuation cases are 100% correct
punctuate (toCap, leadingSpace) (Punc p) = case p of
	Period -> ((True, True), ".")
	Bang -> ((True, True), "!")
	Hyphen -> ((toCap, False), "-")
	Contraction -> ((toCap, False), "'")
	Join -> ((toCap, False), "")
	Comma -> ((toCap, leadingSpace), ",")
	Semicolon -> ((toCap, leadingSpace), ";")
	Colon -> ((toCap, leadingSpace), ":")
	Emdash -> ((toCap, leadingSpace), "--")
	Paragraph -> ((True, True), "\n") -- uh this isn't really intended to work here
punctuate _ _ = error "punctuate given weird values"

capitalize :: String -> String
capitalize [] = []
capitalize (c:rs) = toUpper c : rs

-- removes nils, handles inflection markers, and resolves punctuation streams
resolve :: Maybe (Literal String) -> [Literal String] -> [Literal String]
resolve Nothing [] = []
resolve (Just x) [] = [x]
resolve cur (Nil:rs) = resolve cur rs
resolve cur (Error a:rs) = Error a : resolve cur rs
resolve Nothing (next:rs) = resolve (Just next) rs
resolve (Just cur) (next:rs) = case (cur, next) of
	(Stem w, Inflect i) -> either Error Stem (inflect w i) : resolve Nothing rs
	(Stem w, n) -> Stem w : resolve (Just n) rs
	(Punc p, Punc q) -> resolve (Just $ Punc (max p q)) rs
	(Punc p, n) -> Punc p : resolve (Just n) rs
	(AAn, Stem []) -> Stem "a" : resolve (Just $ Stem []) rs -- i don't know when this would actually happen, but, you know, w/e
	(AAn, Stem (f:w)) -> if f `elem` "aeiouyh" && w /= "uman"
		then Stem "an" : resolve (Just $ Stem (f:w)) rs
		else Stem "a" : resolve (Just $ Stem (f:w)) rs
	(AAn, n) -> AAn : resolve (Just n) rs
	(Inflect i, _) -> Error ("resolution started with inflection " <> show i <> ".") : resolve Nothing (next:rs)
	(Nil, _) -> error "this should be unreachable :O (124)"
	(Error _, _) -> error "this should be unreachable :O (125)"
	-- _ -> error $ "general fallthrough. " <> show (cur, next)

getMatching :: Semantic -> [WordInfo] -> WordInfo
getMatching s ws = fromMaybe (WI 1 s Nothing) . listToMaybe . filter (matches s . sems) $ ws


{- FIXME: okay, some aspect of this (probably this) doesn't work right: required ijs (as seen in the SubjInclude branch here; literally the only path that can return a Nothing) end up being ignored, leading to sentences not failing generation, and instead being generated with missing bits ("You have a."). the commented-out bit in the Compound branch (fail when any components fail, rather than remove any failed components) seems like it would be what would fix the problem, but
	1. that leads to huge chunks of description failing for reasons i'm not clear on and
	2. it still leads to, ultimately, returning a `Just []` result, which ends up getting rendered as " .". which is not ideal.
-}
wordParse :: [WordInfo] -> [Fragment Semantic String] -> WordInfo -> Word String -> Maybe [Literal String]
wordParse _ _ w (Lit l) = Just [l]
wordParse _ _ w (V v) =
	Just $ verbParse v (not . isSingular $ w)
wordParse _ _ w (P p) =
	Just $ pure $ prepParse p
wordParse ws _ w (SubjSingular a c) = case c of
	Nothing -> Just $ if usesArticle w then [Stem a] else []
	Just x -> Just $ if usesArticle (getMatching x ws) then [Stem a] else []
wordParse ws _ w (SubjPlural c) = case c of
	Nothing -> Just $ if isSingular w then [] else [Inflect Plural]
	Just x -> Just $ if isSingular (getMatching x ws) then [] else [Inflect Plural]
wordParse ws _ w (SubjNounPronoun p c) = case (case c of
		Nothing -> pronoun (gender w) p
		Just s -> pronoun (gender $ getMatching s ws) p) of
	"" -> Just []
	v -> Just $ pure . Stem $ v
wordParse ws fs w (SubjInclude req acc f joiner) =
	case (req, wordOutput) of
		(Required, Just []) -> Nothing
		(_, x) -> x
	where
		-- here we recurse through `fragment`, which means if we have a fragment that includes itself (or includes exactly its own semantics) we'll get an infinite loop. the alternative is to not expand fragments interjected into fragments, which mades the system as a whole a lot less useful. so just... don't do that.
		-- (the specific bit is that we are calling wordParse w/ the same fragments, so if we have a fragment that includes itself this will just recurse forever)
		wordOutput :: Maybe [Literal String]
		wordOutput = mconcat $ wordParse ws fs w' <$> fragments
		buh f = show $ (fmap join . sequence $ f, mconcat f)
		fragments :: [Word String]
		fragments =
			  joinInterjections joiner -- intercalated w/ commas (or w/e the joiner says)
			. fmap fragmentWord -- as raw words
			. filter (fragMatch w' f) -- matching fragments
				$ fs
		w' :: WordInfo
		w' = case acc of
			Accumulate -> w
			Discard -> WI 1 None Nothing -- a 0 quantity messes up plurals. (& there's the question of whether we want plurals, or if we want the WordInfo of whatever the matchfragment is, or what -- like, we have replace now; do we ever actually want discard?)
			Replace g -> WI 1 g Nothing
wordParse ws fs w (Compound as) = mconcat $ wordParse ws fs w <$> as
	-- fmap concat $ sequence $ wordParse ws fs w <$> as
wordParse ws fs w (WithFallback initial fallback) =
	case wordParse ws fs w initial of
		Just x -> Just x
		Nothing -> wordParse ws fs w fallback

prepParse :: Preposition -> Literal String
prepParse CommaPrep = Punc Comma
prepParse SemicolonPrep = Punc Semicolon
prepParse ColonPrep = Punc Colon
prepParse (SpecPrep s) = Stem s
prepParse prep = Stem $ toLower <$> show prep

-- this seems to be doing double-duty w/ plural inflection
-- then again some of these are so nonstandard (is/are, etc) that it would be... kind of terrible to have to mark up them all like `Is / s` since uh that's very strange to read
-- this should just use the plural bool to call `inflect`
verbParse :: Verb -> Bool -> [Literal String]
verbParse v plural = pure . Stem $ case v of
	Is -> if' plural "are" "is"
	Has -> if' plural "have" "has"
	Got -> "got"
	Gets -> if' plural "get" "gets"
	Comes -> if' plural "come" "comes"
	Lolls -> if' plural "loll" "lolls"
	Drools -> if' plural "drool" "drools"
	Wrapped -> if' plural "wrap" "wraps"
	Binds -> if' plural "bind" "binds"
	Grows -> if' plural "grow" "grows"
	Frames -> if' plural "frame" "frames"
	Moves -> if' plural "move" "moves"
	Smears -> if' plural "smear" "smears"
	Spills -> if' plural "spill" "spills"
	Slides -> if' plural "slide" "slides"
	Swells -> if' plural "swell" "swells"
	Spans -> if' plural "span" "spans"
	Branches -> if' plural "branch" "branches"
	Bulges -> if' plural "bulge" "bulges"
	Narrows -> if' plural "narrow" "narrows"
	Pushes -> if' plural "push" "pushes"
	Stains -> "stains"
	Pops -> "pop" -- la la wrong plural for demo
	Revolves -> "revolves"
	Sheaths -> "sheaths"
	Unsheaths -> if' plural "unsheath" "unsheaths"
	Inflates -> "inflates"
	Presses -> "presses"
	Stabs -> "stabs"
	Frozen -> "frozen"
	Arc -> "arc"
	Cord -> "cord"
	Forms -> "forms"
	Burns -> "burns"
	Illuminates -> "illuminates"
	Hangs -> "hangs"
	Sends -> "sends"
	Billows -> if' plural "billow" "billows"
	Lurches -> if' plural "lurch" "lurches"
	Spasms -> if' plural "spasm" "spasms"
	Curls -> if' plural "curl" "curls"
	Flexes -> if' plural "flex" "flexes"
	Vanishes -> if' plural "vanish" "vanishes"
	Erupts -> if' plural "erupt" "erupts"
	Streaks -> if' plural "streak" "streaks"
	Counterbalances -> if' plural "counterbalance" "counterbalances"
	Stretches -> if' plural "stretch" "stretches"
	Spreads -> if' plural "spread" "spreads"

if' :: Bool -> a -> a -> a
if' p a b = if p then a else b

{- this function confused me a lot and i wrote it wrong several times.
we have a fragment we want to check for suitability, and we have the context of
	1. the semantic info of the subject we're attached to currently (WordInfo)
	2. a semantic value we want to include here (Semantic)
we want to make sure the fragment fits, meaning we want to make sure all of the fragment's sems can be found within the combination of these two sources. the _fragment_ can have excess semantic tags, but a source having a tag that's not in the fragment is a failed match.
-}
-- discarding the above: this used to take the arguments in the other order (`matches` isn't commutative: one of the args is the template that the other arg attempts to fill), and that was wrong. this might also be wrong, in which case this whole sems concept is incoherent and the problem is structural above and beyond this one function being busted. but hopefully this should fix everything?? we can hope.
fragMatch :: WordInfo -> Semantic -> Fragment Semantic String -> Bool
fragMatch w f (Fragment sem ws) = matches (sems w <> f) sem

gender :: WordInfo -> Gender
gender w = fromMaybe (if isSingular w then It else They) $ mgender w

-- this does lead to some weird things if you have an explicit singular gender but also a > 1 quantity. but when would that situation come up?
-- this also has some issues w/ a quantity of 1 and a plural gender, so maybe those should be separate checks
isSingular :: WordInfo -> Bool
isSingular w = quantity w == 1 && maybe True (not . plural) (mgender w)

usesArticle :: WordInfo -> Bool
usesArticle w = quantity w == 1

fragmentWord :: Fragment x a -> Word a
fragmentWord (Fragment _ w) = w

joinInterjections :: Joiner a -> [a] -> [a]
joinInterjections (normal, twoOnly, final) vs = case vs of
	[a] -> vs
	[a,b] -> a : twoOnly ++ [b]
	rs -> f rs
	where
		f vs = case vs of
			[a, b] -> a : final ++ [b]
			(a:rs) -> a : normal ++ f rs
			v -> v


class SymSem a where
	matches :: a -> Semantic -> Bool
	asSem :: a -> Semantic

{-
matches foo bar = `bar` satisfies all the constraints of `foo`
matches CockOrnament (CockOrnament <> Part Adjective) = True
matches (CockOrnament <> Part Adjective) CockOrnament = False
 -}
instance SymSem Semantic where
	-- this is kind of a mouthful but the idea is we want to make sure each constraint in fs is satisfied at least once by any value in ws
	matches (Multi fs) (Multi ws) = and . fmap (or . (<$> ws) . matches) $ fs
	matches f (Multi ws) = or $ matches f <$> ws
	matches (Multi _) _ = False -- one thing can't satisfy multiple constraints
	matches f w = f == w
	asSem = id

instance SymSem PartOfSpeech where
	matches p = matches (Part p)
	asSem = Part
