{-# LANGUAGE FlexibleInstances #-}

module Description.Parse.Types
  ( Sentence(..)
  , Fragment(..)

  , Diagram(..)
  , d

  , Word(..)
  , Necessary(..)
  , Accumulation(..)
  , Joiner(..)

  , Literal(..)
  , Punctuation(..)

  , WordInfo(..)
  -- i would rather not have these in this file i guess, but it doesn't really matter
  , Verb(..)
  , Preposition(..)
  , Semantic(..)
  , PartOfSpeech(..)
  ) where

import Prelude hiding (Word)

import Data.String (IsString(..))

import Description.Dictionary (Inflection)
import Description.Pronoun (Pronoun, Gender)

import Description.Body.Types (Material(..))

data Sentence def sem word
	= S def (Diagram sem word) [Fragment sem word]
	| PJ (Sentence def sem word) (Fragment sem word) (Sentence () sem word)
	deriving (Eq, Ord, Show, Read)

newtype Diagram sem word = Diagram [Fragment sem word]
	deriving (Eq, Ord, Show, Read)

instance Semigroup (Diagram a b) where
	Diagram a <> Diagram b = Diagram $ a <> b

instance Monoid (Diagram a b) where
	mempty = Diagram []

d :: Fragment sem word -> Diagram sem word
d f = Diagram [f]

data Fragment sem word = Fragment sem (Word word)
	deriving (Eq, Ord, Show, Read)

instance (Semigroup word, Semigroup sem) => Semigroup (Fragment sem word) where
	Fragment sem w <> Fragment sem' w' = Fragment (sem <> sem') (w <> w')

instance (Monoid word, Monoid sem) => Monoid (Fragment sem word) where
	mempty = Fragment mempty mempty



data Word a
	= Lit (Literal a)
	| V Verb
	| P Preposition
	| SubjSingular a (Maybe Semantic)
	| SubjPlural (Maybe Semantic)
	| SubjNounPronoun Pronoun (Maybe Semantic)
	| SubjInclude Necessary Accumulation Semantic (Joiner (Word a))
	| Compound [Word a]
	| WithFallback (Word a) (Word a)
	deriving (Eq, Ord, Show, Read)

instance IsString (Word String) where
	fromString s = Lit $ Stem s

instance Semigroup (Word a) where
	Compound as <> Compound bs = Compound $ as <> bs
	Compound xs <> b
		| null xs = b
		| otherwise = Compound $ xs ++ [b]
	a <> Compound xs
		| null xs = a
		| otherwise = Compound $ a : xs
	a <> b = Compound [a, b]

instance Monoid (Word a) where
	mempty = Compound []

data Necessary = Optional | Required
	deriving (Eq, Ord, Enum, Bounded, Show, Read)

data Accumulation = Accumulate | Discard | Replace Semantic
	deriving (Eq, Ord, Show, Read)

type Joiner a = ([a], [a], [a])



data Literal a
	= Nil
	| AAn -- i can't think of any other syntax constructions that care about the _next_ word
	| Stem a
	| Punc Punctuation
	| Inflect Inflection
	| Error a
	deriving (Eq, Ord, Show, Read)

instance IsString (Literal String) where
	fromString s = Stem s


data Punctuation = Join | Hyphen | Contraction | Comma | Colon | Semicolon | Emdash | Period | Bang | Paragraph
	deriving (Eq, Ord, Show, Read)




data WordInfo = WI
	{ quantity :: Int
	, sems :: Semantic
	, mgender :: Maybe Gender
	}
	deriving (Eq, Show, Read)


-- fixme maybe: plz make all of these a consistent tense, b/c they're kind of a mess to read right now
data Verb
	= Is | Has | Got | Gets
	| Comes
	| Lolls | Drools | Wrapped | Binds
	| Unsheaths | Sheaths | Slides
	| Presses | Stabs | Frozen | Arc | Grows
	| Stains | Smears
	| Spans
	| Pops
	| Revolves
	| Spills | Swells
	| Branches | Bulges | Narrows | Pushes
	| Cord | Forms | Frames | Moves
	| Illuminates | Burns | Hangs | Sends
	| Inflates
	| Billows
	| Lurches | Spasms | Curls | Flexes | Vanishes
	| Erupts | Streaks
	| Counterbalances
	| Stretches | Spreads
	deriving (Eq, Ord, Enum, Bounded, Show, Read)

data Preposition
	= From | Between | In | With | Against | Through | To | Down | Along | Around
	| Behind | Across | Above | Into
	| CommaPrep | SemicolonPrep | ColonPrep
	| SpecPrep String
	deriving (Eq, Ord, Show, Read)


data Semantic
	= None
	| Body
	| Chest
	| Legs | Crotch | Feet
	| OrbForm | LimbFragment
	| Arms
	| Wings
	| Tail
	| Horns | HornPair
	| HornSummary | HornSummaryDependent
	| Cock | CockOrnament | CockTip
	| BallPlacementUnderCock
	| BallSummary
	| Sheath | BallMound | Balls | Knot
	| Asshole
	| Hard | Soft
	| Head | HeadCovering | Halo
	| Eyes | Mouth | Teeth
	| Tongue | TongueContainer
	| Tentacle
	| Interject
--	| MaterialCovering Location
	| Material | AccentMaterial
	| MaterialType Material
	| GenPlural
	| Animal
	| Location | ChimeraPart
	| PreFluid
	| BloodFluid
	| Color
	| Living
	| Count | Size
	| Part PartOfSpeech
	| Multi [Semantic]

	-- sigh i wish i could figure a way to get these / this whole type out of Construction
	| BoneCockIchorCont
	| BoneCockIchor
	deriving (Eq, Ord, Show, Read)

instance Semigroup Semantic where
	None <> b = b
	a <> None = a
	Multi as <> Multi bs = Multi $ as <> bs
	Multi xs <> b
		| null xs = b
		| otherwise = Multi $ b : xs
	a <> Multi xs
		| null xs = a
		| otherwise = Multi $ a : xs
	a <> b
		| a == b = a
		| otherwise = Multi [a, b]
-- both None and Multi [] are effectively mempty, though None will be the result of `mconcat []`
instance Monoid Semantic where
	mempty = None

data PartOfSpeech
	= Subject | Object
	| Noun | Adjective | Verb | Adverb | Pronoun
	| Article
	| Clause
	| Pre | Post
	| Adjunct
	| Preposition | Complement
	| Dependent
	| Listing PartOfSpeech
	deriving (Eq, Ord, Show, Read)

