{-# LANGUAGE ViewPatterns #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE FlexibleInstances #-}

module Server.JSON
	( module Aeson
	, fromResult
	, jsonToEnum

	, jsonToTree
	, mparse
	) where

-- value message instance stuff
import Data.Aeson as Aeson
import Data.Aeson.Types as Aeson (Parser, parse, typeMismatch)
import qualified Data.ByteString.Char8 as B
import qualified Data.ByteString.Lazy.Char8 as BL
import Happstack.Server


-- tfnode stuff
import Description.TfNode hiding (Element(Void))
import Description.Pronoun (Gender(..))

import Data.Char (toLower, toUpper)
import Data.Maybe (listToMaybe)
import Control.Monad (MonadPlus(..), (<=<))

import Data.Text (Text, cons, uncons, unpack)
import Data.Tree

import Data.Vector (toList)
import qualified Data.HashMap.Strict as H (lookup)


import Text.RSS (RSS, rssToXML, showXML)

import Utility.Misc (safeToEnum)

instance ToMessage Value where
	toContentType _ = B.pack "application/json"
	toMessage = encode

instance ToMessage RSS where
	toContentType _ = B.pack "application/rss+xml"
	toMessage = BL.pack . showXML . rssToXML

fromResult :: Result a -> Maybe a
fromResult (Success a) = Just a
fromResult _ = Nothing

jsonToEnum :: (Enum a, Bounded a) => Value -> Maybe a
jsonToEnum = safeToEnum <=< (fromResult . fromJSON)

mparse :: (a -> Parser b) -> a -> Maybe b
mparse p v = case parse p v of
	Success s -> Just s
	_ -> Nothing

jsonToTree :: FromJSON a => Value -> Parser (Tree a)
jsonToTree (Object v) = Node
	<$> v .: "value"
	<*> decodeWith childDecode v "children"
	where
		childDecode = withArray "malformed children" (\arr -> sequence . fmap jsonToTree . toList $ arr)
jsonToTree _ = mzero

-- yes yes orphan instance but the alternative was to make the the tf code directly depend on aeson. (this is still not a great place to put the instances but w/e)
instance ToJSON TfNode where
	toJSON x = toJSON $ nodeLabel x

capitalize :: Text -> Text
capitalize t = case uncons t of
	Nothing -> mempty
	Just (f, rs) -> cons (toUpper f) rs

mread :: Read a => Text -> Maybe a
mread = (\(v, rs) -> if null rs then Just v else Nothing) <=< listToMaybe . reads . unpack . capitalize

instance FromJSON TfNode where
	parseJSON (String s) = case s of
		(mread -> Just r) -> return $ CR r
		(mread -> Just a) -> return $ CA a
		(mread -> Just p) -> return $ CP p
		(mread -> Just s) -> return $ CS s
		(mread -> Just x) -> return $ CX x
		(mread -> Just e) -> return $ CE e
		(mread -> Just w) -> return $ CG (Weapon w)
		(mread -> Just g) -> return $ CG g
		_ -> mzero
	parseJSON _ = mzero

decodeWith :: (Value -> Parser a) -> Object -> Text -> Parser a
decodeWith dec obj key = case H.lookup key obj of
	Nothing -> fail $ "key " ++ show key ++ " not present"
	Just v -> dec v






instance ToJSON Gender where
	toJSON p = toJSON $ toLower <$> show p

instance FromJSON Gender where
	parseJSON (String s) = case s of
		"he" -> return He
		"she" -> return She
		"it" -> return It
		"they" -> return They
		"you" -> return You
		_ -> mzero
	parseJSON _ = mzero
