{-# LANGUAGE OverloadedStrings #-}

module Server.Webserver
	( server
	, Config(..)
	, loadConfig
	) where

-- happstack!
import Happstack.Server hiding (require)
import Network.Socket (Socket)
{-
-- web-routes
import Web.Routes (RouteT, showURL)
import Sites.Hell.Sitemap (SiteRoute(..))

-- web-routes, web-routes-happstack, & web-routes-boomerang
import Web.Routes (Site, RouteT, runRouteT, setDefault)
import Web.Routes.Happstack (implSite)
import Web.Routes.Boomerang (boomerangSite)
-}

-- config handling
import Data.Configurator
import Data.Configurator.Types (Value(..), Configured(..))

import System.Log.Logger
--import System.Log.Handler.Syslog (openlog, Option(..), Facility(..))
import System.Log.Handler.Simple (fileHandler)

-- misc
import System.FilePath.Posix (joinPath)

import Control.Monad (msum, when, mzero)
import Control.Monad.Trans (lift)
import Data.List (isInfixOf)
-- data structures used in Request we need to inspect
import qualified Data.ByteString.Lazy as BS (length)
import qualified Data.ByteString.Char8 as BS (pack)

import qualified Data.Map as M (empty)

import Data.Time (formatTime, defaultTimeLocale)
import Data.Time.Clock (getCurrentTime)

import qualified Sites.Hell.Site as Hell (site)
import qualified Sites.Index.Site as Index (site)

import Server.Layout
import Server.Error

import qualified Text.Blaze.Html5 as H
-- import qualified Text.Blaze.Html5.Attributes as A

server :: Maybe Socket -> Config -> IO ()
server msocket (Config base host port localport asDaemon) = do
	let debug = not asDaemon
	startTime <- getCurrentTime
	logLn INFO $ concat
		[ "** starting server "
		, host
		, ":"
		, show port
		, " on "
		, formatTime defaultTimeLocale "[%d/%b/%Y:%H:%M:%S %z]" startTime
		]
	let conf = Conf port Nothing (Just logMAccess) 30 Nothing
	let serveHTTP = case msocket of
		Nothing -> simpleHTTP conf
		Just socket -> simpleHTTPWithSocket socket conf
	serveHTTP $ (do
		-- | that's 0 bytes for temp files (from e.g., forms), 1 MB (1048576 bytes) for other header data (e.g., http headers, form post data, cookies, query info), and 10000 bytes for "overhead for headers in multipart/form-data", which idk what that is :V
		decodeBody (defaultBodyPolicy "/tmp/" 0 1048576 10000)
		setSize <$> msum
			[ Index.site base host localport ""
			-- it'd be nice to get this (the error-handling line/function) out of server.webserver and into sites.hell.site but given what the error handler has to do that doesn't seem feasible
			, basicErrorHandler debug `except` Hell.site base host localport "/hell"
			, dir "static"
				$ serveDirectory DisableBrowsing [] $ joinPath [base, "static"]
			, asResponseCode 404 <$> page404
			]) `catch` logCatch debug

logCatch :: Bool -> SomeException -> ServerPart Response
logCatch debug ex = do
	lift $ logLn WARNING $ concat
		[ "*** error on next line: "
		, show ex
		]
	return . asResponseCode 500
		$ errorPage 500 $ do
			H.p "SOMETHING WENT WRONG"
			if debug
				then
					H.p $ toMarkup $ show ex
				else do
					H.p "this is probably fixable"
					H.p "plz contact an admin about what happened (that's xax)"

-- for hell-specific errors
basicErrorHandler :: Bool -> HellError -> ServerPartT IO Response
basicErrorHandler _ NoError = mzero
basicErrorHandler debug (HellError err) = do
	lift $ logLn WARNING $ concat
		[ "*** error on next line: "
		, show err
		]
	return . asResponseCode 500
		$ errorPage 500 $ do
			H.p "SOMETHING WENT WRONG"
			if debug
				then
					sequence_ $ (H.p . toMarkup) <$> err
				else do
					H.p "this is probably fixable"
					H.p "plz contact an admin about what happened (that's xax)"

-- i have never got the things like notFound or internalServerError to actually give the right response code, so i'm doing it manually here
asResponseCode :: ToMessage a => Int -> a -> Response
asResponseCode code resp = setHeaderBS "Content-Type" (toContentType resp)
	$ Response code M.empty nullRsFlags (toMessage resp) Nothing

-- ...likewise happstack has _gotta_ do this itself; there is no way the user is expected to specify response sizes
setSize :: Response -> Response
setSize resp = setHeaderBS "Content-Length" (BS.pack . show $ responseSize resp) resp

responseSize :: Response -> Integer
responseSize r = case r of
	(Response _ _ _ b _) -> fromIntegral $ BS.length b
	(SendFile _ _ _ _ path _ c) -> c

data Config = Config
	{ base :: String
	, host :: String
	, port :: Int
	, redirPort :: Int
	, asDaemon :: Bool
	}

data DaemonStatus = Standalone | Daemon
	deriving (Eq, Ord, Enum, Bounded)

instance Configured DaemonStatus where
	convert (String s)
		| s == "standalone" = Just Standalone
		| s == "daemon" = Just Daemon
	convert _ = Nothing

loadConfig :: String -> IO Config
loadConfig config = do
	config <- load [Required config]
	base <- require config "base"
	host <- require config "host"
	port <- require config "port"
	redirPort <- lookupDefault port config "redir_port"
	runtype <- require config "runtype"
	accesslog <- require config "access"
	errorlog <- require config "error"
	{-
	-- i could never figure out how this worked or what if anything it was logging to; i don't know how syslog works or how this interfaces with it.
	when (runtype == Daemon) $ do
		logHandler <- openlog "HELL SERVER" [PID] USER DEBUG
		updateGlobalLogger "Happstack.Server.AccessLog.Combined" (addHandler logHandler)
	-}
	accessLog <- fileHandler accesslog DEBUG
	errorLog <- fileHandler errorlog WARNING
	updateGlobalLogger "Happstack.Server.AccessLog.Combined" (addHandler accessLog)
	updateGlobalLogger "Happstack.Server.AccessLog.Combined" (addHandler errorLog)
	-- i think what this does is make the happstack log output to stdout
	updateGlobalLogger "Happstack.Server.AccessLog.Combined" (setLevel DEBUG)
	return $ Config base host port redirPort (runtype == Daemon)


-- todo in the future: change type to `RouteT SiteRoute (ServerPartT IO) Html`, so i can use type-safe links
page404 :: ServerPartT IO Html
page404 = do
	req <- askRq
	let splashUrl = "/hell" :: String
	return $
		H.docTypeHtml $ do
			H.head $ H.title "Nothing here."
			H.p $ do
				"Nothing at "
				H.code $ toHtml $ rqUri req
				". Maybe try the "
				link splashUrl ("index" :: String)
				"?"
			when (isInfixOf "%3C" $ rqUri req) $
				H.p "also nice hax. gj. well done."
{-
-- but i have no clue how to actually use this? like we put it into a response but i don't really know how to 'run' the route transformer exactly.
page404' :: RouteT SiteRoute (ServerPartT IO) Html
page404' = do
	splashUrl <- showURL Splash
	req <- askRq
	return $
		H.docTypeHtml $ do
			H.head $ H.title "Nothing here."
			H.p $ do
				"Nothing at "
				H.code $ toHtml $ rqUri req
				". Maybe try the "
				link splashUrl ("index" :: String)
				"?"
			when (isInfixOf "%3C" $ rqUri req) $
				H.p "also nice hax. gj. well done."
-}
