{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE BangPatterns #-}

module Sites.Hell.Routes.DBCheck
	( dbCheckPage
	) where

{-
db admin stuff:

- flag and unflag comments
- delete comments
- mark comments as being by an admin
- post news posts
- edit news posts

- set comment style to moderated (all new comments flagged by default)
- disable new comments

- comments should log ip!! once that means something

-}

import Control.Monad
import Control.Monad.Trans
import Control.DeepSeq

import qualified Data.Map as Map

import Web.Routes (RouteT) -- showURL

import Sites.Hell.Sitemap (SiteRoute(..))

import Happstack.Server (ServerPartT, Response(..), toResponse)
import Server.Error (HellServer(..), except, asExcept, catch, ExceptT(..))
import Server.Layout
import qualified Text.Blaze.Html5 as H
import qualified Text.Blaze.Html5.Attributes as A

import Server.Database

type DBServer = ServerPartT (ExceptT SqlError IO)

dbCheckPage :: RouteT SiteRoute HellServer Response
dbCheckPage = lift . asExcept $ (dbErrorHandler False `except` dbCheckPage') `catch` dbErrorHandler True

dbCheckPage' :: ServerPartT (ExceptT SqlError IO) Response
dbCheckPage' = do
	(foo, posts, comments) <- withHellDB $ (\conn -> do
			let foo = connectionDebug conn

			posts <- describeTable conn "newspost"
			comments <- describeTable conn "newspost_comment"
			return (foo, posts, comments)
		)
	return $ force $ toResponse $ minimal "HELL GAME &rarr; db check"
		(do
			stylesheet "/static/css/main.css"
			stylesheet "/static/css/analysis.css"
		)
		(do
			H.div ! A.id "main" $ do
				H.p $ "db connection successful"
				H.p $ do
					"db type: "
					toHtml foo
				H.p $ do
					"status of " >> H.kbd "newspost" >> " table: "
					newspostValid posts
				H.p $ do
					"status of " >> H.kbd "newspost_comment" >> " table: "
					newspostCommentValid comments
		)

newspostValid :: [(String, SqlColDesc)] -> Html
newspostValid [] = H.span ! A.class_ "error" $ "no table"
newspostValid l = let m = Map.fromList l
	in case (Map.lookup "id" m, Map.lookup "body" m, Map.lookup "date" m, length l) of
		(Just id, Just body, Just date, 4) -> "probably okay"
		_ -> "damaged"

newspostCommentValid :: [(String, SqlColDesc)] -> Html
newspostCommentValid [] = H.span ! A.class_ "error" $ "no table"
newspostCommentValid l = let m = Map.fromList l
	in case (Map.lookup "id" m, Map.lookup "newspostid" m, Map.lookup "author" m, Map.lookup "body" m, Map.lookup "hidden" m, Map.lookup "date" m, length l) of
		_ -> "who knows"
{-
CREATE TABLE newspost
	( id serial PRIMARY KEY
	, body text NOT NULL
	, hilight bool NOT NULL DEFAULT false
	, date timestamp NOT NULL DEFAULT now()
	);
CREATE TABLE newspost_comment
	( id serial PRIMARY KEY
	, newspostid integer REFERENCES newspost (id)
	, author text NOT NULL DEFAULT 'anonymous'
	, body text NOT NULL
	, hilight bool NOT NULL DEFAULT false
	, hidden bool NOT NULL DEFAULT false
	, date timestamp NOT NULL DEFAULT now()
	);
-}

connectionDebug :: IConnection a => a -> String
connectionDebug conn = concat
	[ "[",hdbcDriverName conn
	, "] ", hdbcClientVer conn
	, " // ", dbServerVer conn
	, " ", if dbTransactionSupport conn then "with transactions" else "without transactions"
	]

-- for db errors
dbErrorHandler :: Bool -> SqlError -> ServerPartT IO Response
dbErrorHandler caught (SqlError "0000" _ _) = mzero
dbErrorHandler caught (SqlError errCode native errorMsg) = do
	let except = if caught then "thrown SQL error" else "stored SQL exception"
	return $ toResponse $ minimal "HELL GAME &rarr; db check"
		(do
			stylesheet "/static/css/main.css"
			stylesheet "/static/css/analysis.css"
		)
		(do
			H.div ! A.id "main" $ do
				H.p $ "DB check failed."
				H.p $ do
					"Caught "
					except
					" during check. (code "
					H.span ! A.class_ "error" $ toHtml native
					when (errCode /= "") $ do
						"/"
						H.span ! A.class_ "error" $ toHtml errCode
					")"
				H.p $ "The error was:"
				H.p ! A.class_ "error" $ toHtml errorMsg
		)
