
function Point2d (x, y) {
	this.x = x;
	this.y = y;
}
Point2d.prototype =
	{ magnitude: function (type) {
			if (type === "manhattan") {
				return Math.abs (this.x) + Math.abs (this.y);
			}
			return Math.sqrt (this.x * this.x + this.y * this.y);
		}
	, distance: function (pt) {
			return this.subtract (pt).magnitude();
		}
	, angle: function () {
			return Math.atan2 (this.x, this.y);
		}
	, rotate: function (rad) {
			var x_ = this.x * Math.cos(rad) - this.y * Math.sin(rad);
			var y_ = this.y * Math.cos(rad) + this.x * Math.sin(rad);
			return new Point2d (x_, y_);
		}
	, normalize: function () {
			var m = this.magnitude();
			if (m === 0) {
				return new Point2d (0, 0);
			}
			return new Point2d (this.x / m, this.y / m);
		}
	, equal: function (pt) {
			return this.x === pt.x && this.y === pt.y;
		}
	, equals: function (pt) {
			return this.equal (pt);
		}
	, plus: function (pt) {
			return this.add(pt);
		}
	, minus: function (pt) {
			return this.subtract(pt);
		}
	, add: function (pt) {
			return new Point2d (this.x + pt.x, this.y + pt.y);
		}
	, subtract: function (pt) {
			return new Point2d (this.x - pt.x, this.y - pt.y);
		}
	, timesScalar: function (s) {
			return new Point2d (this.x * s, this.y * s);
		}
	, inBounds: function (lowX, hiX, lowY, hiY) {
			if (lowY === undefined) {
				lowY = lowX;
			}
			if (hiY === undefined) {
				hiY = hiX;
			}
			return this.x >= lowX && this.x < hiX &&
				this.y >= lowY && this.y < hiY;
		}
	, dot: function (vec) {
			return this.x * vec.x + this.y * vec.y;
		}
	, aa: function () { // does this represent an axis-aligned vector
			return this.x === 0 || this.y === 0;
		}
	, adjacent: function () {
		return [ new Point2d (this.x + 1, this.y)
			, new Point2d (this.x - 1, this.y)
			, new Point2d (this.x, this.y + 1)
			, new Point2d (this.x, this.y - 1)
			];
		}
	, toString: function () {
			return "{" + this.x + "," + this.y + "}";
		}
	};

function Point3d (x, y, z) {
	this.x = x;
	this.y = y;
	this.z = z;
}
Point3d.prototype =
	{ magnitude: function () {
			return Math.sqrt (this.x * this.x + this.y * this.y + this.z * this.z);
		}
	, add: function (pt) {
			return new Point3d (this.x + pt.x, this.y + pt.y, this.z + pt.z);
		}
	, subtract: function (pt) {
			return new Point3d (this.x - pt.x, this.y - pt.y, this.z - pt.z);
		}
	// haven't tested this; don't know how well it works
	, toSpherical: function () {
			// r is radius
			var r = this.magnitude();
			// theta is inclination/altitude; between 0 and π radians. 0 is north pole, π is south pole
			var theta = this.z === 0 ? 0 : Math.acos (this.z / r);
			// phi is azimuth; between -π and π radians. 0 is north, i think
			var phi = this.x === 0 ? 0 : Math.atan2 (this.y, this.x);
			return new Spherical (r, theta, phi);
		}
	, negate: function () {
			return new Point3d (-this.x, -this.y, -this.z);
		}
	, normalize: function () {
			var m = this.magnitude();
			return new Point3d (this.x / m, this.y / m, this.z / m);
		}
	, timesScalar: function (s) {
			return new Point3d (this.x * s, this.y * s, this.z * s);
		}
	, cross: function (vec) {
			return new Point3d
				( this.y * vec.z - this.z * vec.y
				, this.z * vec.x - this.x * vec.z
				, this.x * vec.y - this.y * vec.x
				);
		}
	, dot: function (vec) {
			return this.x * vec.x + this.y * vec.y + this.z * vec.z;
		}
	, rotate: function (axis, theta) {
			var matrix = Matrix44.prototype.axisAngleRotationMatrix (axis, theta);
			return matrix.multiplyVector (this);
		}
	// return a rotation function that maps `this -> align` and performs the corresponding rotation for all other vectors
	, alignFunc: function (align) {
			var axis = this.cross (align).normalize();
			var theta = Math.acos (this.normalize().cross (align.normalize()));
			var matrix = Matrix44.prototype.axisAngleRotationMatrix (axis, theta);
			return function (vec) {
				return matrix.multiplyVector (vec);
			}
		}
	};

function Matrix44 (vs) {
	var i = 0;
	this.vs =
		[ 1, 0, 0, 0
		, 0, 1, 0, 0
		, 0, 0, 1, 0
		, 0, 0, 0, 1
		];
	if (vs) {
		for (i = 0; i < 16; i++) {
			this.vs[i] = vs[i];
		}
	}
}
Matrix44.prototype =
	{ axisAngleRotationMatrix: function (angle, theta) {
			var nAxis = axis.normalize();
			var x = nAxis.x;
			var y = nAxis.y;
			var z = nAxis.z;
			var c = Math.cos (theta);
			var s = Math.sin (theta);
			var t = 1 - c;
			var matrix = new Matrix44 (
				[ t * x * x + c    , t * x * y + z * s, t * x * z - y * s, 0
				, t * x * y - z * s, t * y * y + c    , t * y * z + x * s, 0
				, t * x * z + y * s, t * y * z - x * s, t * z * z + c    , 0
				, 0                , 0                , 0                , 1
				]);
			return matrix;
		}
	, matrixMultiply: function (m44) {
			var r = new Matrix44();
			var c = 0;
			var i;
			var j;
			for (c = 0; c < 16; c++) {
				i = Math.floor (c / 4);
				j = c % 4;

				c++;
			}
			r[0]
				= this.vs[0] * m44.vs[0]
				+ this.vs[1] * m44.vs[4]
				+ this.vs[2] * m44.vs[8]
				+ this.vs[3] * m44.vs[12];
			r[1]
				= this.vs[0] * m44.vs[1]
				+ this.vs[1] * m44.vs[5]
				+ this.vs[2] * m44.vs[9]
				+ this.vs[3] * m44.vs[13];
			r[2]
				= this.vs[0] * m44.vs[2]
				+ this.vs[1] * m44.vs[6]
				+ this.vs[2] * m44.vs[10]
				+ this.vs[3] * m44.vs[14];
			r[3]
				= this.vs[0] * m44.vs[3]
				+ this.vs[1] * m44.vs[7]
				+ this.vs[2] * m44.vs[11]
				+ this.vs[3] * m44.vs[15];
			r[4]
				= this.vs[4] * m44.vs[0]
				+ this.vs[5] * m44.vs[4]
				+ this.vs[6] * m44.vs[8]
				+ this.vs[7] * m44.vs[12];
			// ...
			return r;
		}
	, multiplyVector: function (vec3) {
			var x
				= this.vs[0] * vec3.x
				+ this.vs[1] * vec3.y
				+ this.vs[2] * vec3.z
				+ this.vs[3] * 0;
			var y
				= this.vs[4] * vec3.x
				+ this.vs[5] * vec3.y
				+ this.vs[6] * vec3.z
				+ this.vs[7] * 0;
			var z
				= this.vs[8] * vec3.x
				+ this.vs[9] * vec3.y
				+ this.vs[10] * vec3.z
				+ this.vs[11] * 0;
			return new Point3d (x, y, z);
		}
	};

/*
function Quaternion (w, x, y, z) {
	this.w = w;
	this.x = x;
	this.y = y;
	this.z = z;
}
Quaternion.prototype =
	{

	};
*/

function linearize (dim, x, y) {
	return y * dim + x;
}
function unlinearize (dim, i) {
	return new Point2d (i % dim, (i / dim)|0);
}


// theta is inclination/altitude; between 0 and π radians. 0 is north pole, π is south pole
// phi is azimuth; between -π and π radians. 0 is north, i think
function Spherical (r, theta, phi) {
	this.r = r;
	this.theta = theta;
	this.phi = phi;
}
Spherical.prototype =
	// haven't tested this; don't know how well it works
	{ toCartesian: function () {
			var x = this.r * Math.sin (this.theta) * Math.cos (this.phi);
			var y = this.r * Math.sin (this.theta) * Math.sin (this.phi);
			var z = this.r * Math.cos (this.theta);
			return new Point3d (x, y, z);
		}
	};
