
/* Graph assumptions:
	* edges are bidirectional
	* multiedges are fine
	* .getNodeId will get the FIRST node with that label, if there are duplicate nodes
	* .minimumSpanningTree returns null for disconnected graphs, not a spanning forest
*/
function Graph (nodes, edges) {
	if (!nodes) {
		nodes = [];
	}
	if (!edges) {
		edges = [];
	}
	this.nodes = nodes;
	this.edges = edges;
	this.pathInfo = {};
}
Graph.prototype =
	{ addNode: function (label) {
			this.nodes.push (label);
		}
	, getNodeId: function (nodeLabel) {
			var match = null;
			this.nodes.map (function (l, i) {
				if (match === null && l === nodeLabel) {
					match = i;
				}
			});
			//console.log ("match: \"" + match + "\"");
			return match;
		}
	, addEdge: function (n, m, label) {
			this.edges.push ({n: n, m: m, label: label});
		}
	, edgesBetween: function (n, m) {
			return this.edges.filter (function (e) {
				return (e.n === n && e.m === m) || (e.m === n && e.n === m);
			});
		}
	, edgesMatching: function (n, m, pred) {
			return this.edgesBetween (n, m).filter (pred);
		}
	, adjacent: function (n) {
			return this.edges.filter (function (e) {
				return (e.n === n || e.m === n);
			});
		}

	, dijkstras: function (start) {
			var
				open = [start],
				info = {},
				neighbors,
				neighbor, i,
				active, activeInfo, neighborInfo;
			info[start] =
				{ length: 0
				, parent: null
				};
			//console.log ("running dijkstras (at " + start + ")");
			while ((active = open.shift()) !== undefined) {
				//console.log ("active: " + active);
				neighbors = this.adjacent (active).map (function (neighbor) {
					return neighbor.n === active ? neighbor.m : neighbor.n;
				});
				//console.log (neighbors);
				neighbors.map (function (neighborId) {
					//console.log ("got " + neighborId);
					if (!info[neighborId]) {
						info[neighborId] =
							{ length: info[active].length + 1
							, parent: active
							};
						//console.log ("new node; pushing to open");
						open.push (neighborId);
					} else if (info[neighborId].length > info[active].length + 1) {
						info[neighborId].length = info[active].length + 1;
						info[neighborId].parent = active;
					}
				});
			}
			//console.log ("done w/ dijkstras; " + Object.keys(info).length + " nodes reachable");
			this.pathInfo = info;
			return info;
		}
	, pathTo: function (end) {
			if (!this.pathInfo) {
				throw new Error ("graph: can't make path to " + end + " without running dijkstras to begin with");
			}
			var path = [end];
			var active = end;
			if (this.nodes[active] === undefined) {
				throw new Error ("graph: no node \"" + active + "\" in graph; can't pathfind");
			} else if (!this.pathInfo[active]) {
				console.log (this);
				throw new Error ("graph: no path to node \"" + active + "\" in graph");
			}
			//console.log (this.pathInfo);
			//console.log (active);
			while (this.pathInfo[active].parent !== null) {
				path.unshift (this.pathInfo[active].parent);
				active = this.pathInfo[active].parent;
				//console.log (active);
			}
			return path;
		}
	, minimumSpanningTree: function (edgeWeightFromLabel, startingNode) {
			var mst = [];
			var usedNodes = [];
			var edges;
			var nextEdges;
			var nextEdge;
			if (this.nodes.length === 0) {
				return mst;
			}
			if (startingNode === undefined) {
				startingNode = 0;
			}
			usedNodes.push (startingNode);
			while (usedNodes.length < this.nodes.length) {
				nextEdges = this.edges.filter (function (e) {
					return (usedNodes.indexOf (e.n) !== -1 && usedNodes.indexOf (e.m) === -1) ||
						(usedNodes.indexOf (e.m) !== -1 && usedNodes.indexOf (e.n) === -1);
				});
				if (nextEdges.length === 0) {
					console.error ("graph is disconnected; can't make spanning tree");
					return null;
				}
				nextEdges = nextEdges.reduce (function (shortests, e) {
					var eWeight = edgeWeightFromLabel (e.label);
					var aWeight = shortests.length === 0 ? 9999 : edgeWeightFromLabel (shortests[0].label);
					if (eWeight < aWeight) {
						return [e];
					} else if (eWeight === aWeight) {
						shortests.push(e);
						return shortests;
					}
					return shortests;
				}, []);
				nextEdge = randomFrom (nextEdges);
				mst.push (nextEdge);
				usedNodes.push (usedNodes.indexOf (nextEdge.n) === -1 ? nextEdge.n : nextEdge.m);
			}
			return mst;
		}
	, resetEdges: function (edges) {
			if (edges === undefined || edges === null) {
				this.edges = [];
			} else {
				this.edges = edges;
			}
		}
	};
